# system modules
import logging
import urllib
import json
import functools
import re

# internal modules
from sensemapi import paths
from sensemapi.reprobject import ReprObject
from sensemapi.senseBox import senseBox
from sensemapi.errors import *
from sensemapi.utils import *

# external modules
import requests

logger = logging.getLogger(__name__)

class SenseMapClient(ReprObject):
    """
    Client to interface the `OpenSenseMap API <https://api.opensensemap.org>`_
    """
    def _get_box(self, id, format = None):
        """
        Issue the request to retreive a single senseBox

        Args:
            id (str) : the senseBox id to retreive
            format (str, optional): one of ``"json"`` and ``"geojson"``

        Returns:
            dict : the API response
        """
        response = requests.get(
            urllib.parse.urljoin(paths.BOXES+"/",id),
            params = {"format":format} if format else {}
            )
        response_json = response.json()
        logger.debug("API responded JSON:\n{}".format(
            json.dumps(response_json, sort_keys = True, indent = 4)))
        if response.status_code == 200:
            return response_json
        else:
            message = response_json.get("message")
            raise OpenSenseMapAPIError("Could not retreive with id '{}'{}"
                .format(id, ": {}".format(message) if message else ""))

    def get_box(self, id):
        """
        Retreive one :any:`senseBox`

        Args:
            id (str) : the senseBox id to retreive

        Returns:
            senseBox : the retreived senseBox
        """
        box = senseBox.from_json(self._get_box(id = id, format = "json"))
        box.client = self
        return box

    def _post_measurement(self, box_id, sensor_id, value, time = None,
        lat = None, lon = None, height = None):
        """
        Issue a request to upload a new measurement

        Args:
            box_id (str) : the senseBox id
            sensor_id (str) : the sensor's id
            value (float) : the current measurement value
            time (datetime.datetime, optional) : the time of the measurement
            lat, lon, height (float,optional) : the current position

        Returns:
            True : on success
        """
        assert box_id is not None, "box_id must  be defined"
        assert sensor_id is not None, "sensor_id must  be defined"
        d = {}
        d["value"] = float(value)
        if time:
            d["createdAt"] = date2str(time)
        try:
            d["location"] = location_dict(lat, lon, height)
        except ValueError:
            pass
        logger.debug("Sending Request with JSON:\n{}"
            .format(pretty_json(d)))
        response = requests.post(
            functools.reduce(urllib.parse.urljoin,
                [paths.BOXES+"/",box_id+"/",sensor_id]),
            json = d,
            )
        response_json = response.json()
        logger.debug("API responded JSON:\n{}"
            .format(pretty_json(response_json)))
        if hasattr(response_json, "get"): # is a dict
            message = response_json.get("message")
            raise OpenSenseMapAPIError("Posting measurement didn't work{}"
                .format(": "+ message or ""))
        else: # no dict
            if re.search("measurement\s+saved\s+in\s+box",response_json):
                return True

    def post_measurement(self, sensor):
        """
        Upload the current measurement of a given :any:`senseBoxSensor`.

        Args:
            sensor (senseBoxSensor) : the sensor
        """
        assert sensor.id, "the given sensor does not have an id"
        assert sensor.box, "the given sensor does not know its senseBox"
        assert sensor.box.id, "the given sensor's senseBox does not have an id"
        assert sensor.last_value, "the given sensor does not have a last_value"
        post_kwargs = {}
        post_kwargs.update(
            box_id = sensor.box.id,
            sensor_id = sensor.id,
            value = float(sensor.last_value),
            )
        if sensor.box.exposure == "mobile":
            post_kwargs.update(
                lat = box.current_lat,
                lon = box.current_lon,
                height = box.current_height,
                )
        if sensor.last_time:
            post_kwargs.update(time = sensor.last_time)
        return self._post_measurement(**post_kwargs)



class OpenSenseMapAPIError(Exception):
    """
    Base Exception class raised whenever something is strange
    """
    pass

class OpenSenseMapAPIResponseError(Exception):
    """
    Exception raised when the API responded something strange
    """
    pass

class OpenSenseMapAPIAuthenticationError(OpenSenseMapAPIError):
    """
    Exception raised when the authentication failed
    """
    pass

class OpenSenseMapAPIInvalidCredentialsError(
    OpenSenseMapAPIAuthenticationError):
    """
    Exception raised when the login failed due to invalid credentials
    """
    pass

class OpenSenseMapAPIOutdatedTokenError(OpenSenseMapAPIAuthenticationError):
    """
    Exception raised when a token is outdated
    """
    pass

class OpenSenseMapAPIPermissionError(OpenSenseMapAPIAuthenticationError):
    """
    Exception raised when the account does not have certain permissions
    """
    pass

class NoClientError(Exception):
    pass

class ConfirmationError(Exception):
    pass



SIGN_IN = "https://api.opensensemap.org/users/sign-in"
SIGN_OUT = "https://api.opensensemap.org/users/sign-out"
REFRESH = "https://api.opensensemap.org/users/refresh-auth"
USER_DETAILS =  "https://api.opensensemap.org/users/me"
BOXES =  "https://api.opensensemap.org/boxes"



# system modules
import logging
import datetime
import inspect
import iso8601
import json

# internal modules

# external modules

OPENSENSEMAP_DATETIME_FORMAT_UTC = "%Y-%m-%dT%H:%M:%S.000Z"

def str2date(s):
    """
    Convert a string in RFC3339 format to a Python :any:`datetime.datetime`
    object.

    Args:
        s (str) : the string to convert. Needs to be in RFC3339 format.

    Returns:
        datetime.datetime : the converted date
    """
    return iso8601.parse_date(s)

def date2str(d):
    """
    Stringify a :any:`datetime.datetime` object to RFC3339 format

    Args:
        d (datetime.datetime) : the datetime to convert

    Returns:
        str : the converted string
    """
    try:
        utc = d.astimezone(datetime.timezone.utc)
    except ValueError:
        utc = d.replace(tzinfo = datetime.timezone.utc)
    return utc.strftime(OPENSENSEMAP_DATETIME_FORMAT_UTC)

def pretty_json(d): # pragma: no cover
    """
    Pretty-format a JSON dict

    Args:
        d (dict) : the JSON dict

    Returns:
        str : the formatted string
    """
    return json.dumps(d, sort_keys = True, indent = 4)

def location_dict(lat = None, lon = None, height = None):
    """
    Create a location dict from the single values

    Args:
        lat, lon, height (float, optional) : the position

    Returns:
        dict : the location dict

    Raises:
        ValueError : if not enough values are specified
    """
    if lat is not None and lon is not None:
        if height is not None:
            return {"lat":lat, "lng":lon, "height":height}
        else:
            return {"lat":lat, "lng":lon}
    else:
        raise ValueError("At least 'lat' and 'lon' need to be defined")

def simplegetter(fn): # pragma: no cover
    """
    Property getter method decorator for easy getter setup. Getter methods
    decorated with this decorator should return the default value. When the
    property is requested, it is checked whether the internal attribute
    prefixed with ``_`` exists and if so it is returned. If not, it is set
    to the return value of the decorated method and then it is returned.

    Parameters
    ----------

    Returns
    -------

    callable
        The decorated callable
    """
    propname = fn.__name__
    attrname = "_{}".format(propname)
    try:
        inspect.getfullargspec(fn)[0]
    except KeyError:  # pragma: no cover
        raise ValueError(
            "`simplegetter` decorator can only be used for methods that take "
            "the object reference as first argument"
        )

    @property
    def getter(self):
        try:
            return getattr(self, attrname)
        except AttributeError:
            fnval = fn(self)
            setattr(self, attrname, fnval)
        return getattr(self, attrname)

    getter.__doc__ = fn.__doc__
    return getter


def simplesetter(prop, del_on_exceptions=()): # pragma: no cover
    """
    Property setter method decorator for easy setter setup. Setter methods
    decorated with this decorator should take the object reference and the new
    value as argument, modify the value as desired and return it. The return
    value will be stored in the internal attribute prefixed with
    ``_``.

    Parameters
    ----------

    prop : property
        the property
    del_on_exceptions : sequence of BaseException subclasses
        delete the internal attribute if any of these in
        :any:`exceptions` occur

    Returns
    -------

    callable
        The decorated callable
    """
    def simplesetter_decorator(fn):
        propname = fn.__name__
        attrname = "_{}".format(propname)
        try:
            inspect.getfullargspec(fn)[0]
            inspect.getfullargspec(fn)[1]
        except KeyError:  # pragma: no cover
            raise ValueError(
                "`simplegetter` decorator can only be used for methods that "
                "take the object reference as first argument and the new "
                "property value as second argument"
            )

        def setter(self, newval):
            try:
                converted = fn(self, newval)
                setattr(self, attrname, converted)
            except del_on_exceptions:
                try:
                    delattr(self, attrname)
                except AttributeError:
                    pass

        setter.__doc__ = fn.__doc__
        setter = prop.setter(setter)
        return setter

    return simplesetter_decorator



# system modules
import logging
import inspect
import collections
import itertools

# internal modules
from sensemapi.reprobject import ReprObject
from sensemapi.sensor import senseBoxSensor, senseBoxSensorCollection
from sensemapi import paths
from sensemapi.errors import *
from sensemapi.utils import *

# external modules
import requests

logger = logging.getLogger(__name__)

class senseBox(ReprObject):
    """
    Class representation of a senseBox
    """
    def __init__(self,
        id = None,
        exposure = None,
        grouptag = None,
        description = None,
        name = None,
        sensors = None,
        image = None,
        current_lat = None,
        current_lon = None,
        current_height = None,
        weblink = None,
        ):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @simplegetter
    def id(self):
        return None

    @simplesetter(id)
    def id(self, new):
        return new

    @simplegetter
    def exposure(self):
        return None

    @simplesetter(exposure)
    def exposure(self, new):
        return new

    @simplegetter
    def grouptag(self):
        return None

    @simplesetter(grouptag)
    def grouptag(self, new):
        return new

    @simplegetter
    def description(self):
        return None

    @simplesetter(description)
    def description(self, new):
        return new

    @simplegetter
    def name(self):
        return None

    @simplesetter(name)
    def name(self, new):
        return new

    @property
    def sensors(self):
        try:
            self._sensors
        except AttributeError:
            self._sensors = senseBoxSensorCollection()
            self._sensors.box = self
        return self._sensors

    @sensors.setter
    def sensors(self, new):
        if hasattr(new, "sensors"): # is already a senseBoxSensorCollection
            self._sensors = new
        else:
            self._sensors = senseBoxSensorCollection()
            self._sensors.box = self
            self._sensors.sensors = new

    @simplegetter
    def image(self):
        return None

    @simplesetter(image)
    def image(self, new):
        return new

    @simplegetter
    def weblink(self):
        return None

    @simplesetter(weblink)
    def weblink(self, new):
        return new

    @simplegetter
    def client(self):
        return None

    @simplesetter(client)
    def client(self, new):
        return new

    @simplegetter
    def current_lat(self):
        return None

    @simplesetter(current_lat)
    def current_lat(self, new):
        return new

    @simplegetter
    def current_lon(self):
        return None

    @simplesetter(current_lon)
    def current_lon(self, new):
        return new

    @simplegetter
    def current_height(self):
        return None

    @simplesetter(current_height)
    def current_height(self, new):
        return new

    @property
    def location(self):
        """
        A location object as used by the API

        :type: :any:`dict`
        """
        return location_dict(
            self.current_lat, self.current_lon, self.current_height)

    def new_sensor(self, *args, **kwargs):
        """
        Add a new :any:`senseBoxSensor` to this :any:`senseBox`

        Args:
            args, kwargs : arguments passed to :any:`senseBoxSensor.__init__`

        Returns:
            senseBoxSensor : the newly created sensor
        """
        if self.sensors is None:
            self.sensors = []
        sensor = senseBoxSensor(*args, **kwargs)
        self.add_sensor(sensor)
        return sensor

    def add_sensor(self, sensor):
        """
        Add a :any:`senseBoxSensor` to this :any:`senseBox`

        Args:
            sensor (senseBoxSensor) : the sensor to add
        """
        if self.sensors is None:
            self.sensors = []
        self.sensors.append(sensor)

    def update_from_json(self, d):
        """
        Create a :any:`senseBox` from a :any:`dict`.

        Args:
            d (dict): the dictionary to update this :any:`senseBox` from
        """
        coords = d.get("currentLocation",{}).get("coordinates")
        lat, lon, height = [None] * 3
        try:
            lon, lat, height = coords
        except (TypeError, ValueError):
            try:
                lon, lat = coords
            except (TypeError, ValueError): # pragma: no cover
                pass
        if self.sensors is None:
            self.sensors = []
        for sensor_json in d.get("sensors",[]):
            sensor_json_copy = sensor_json.copy()
            sensor_id = sensor_json_copy.get("_id")
            own_sensor = self.sensors.by_id.get(sensor_id)
            if own_sensor:
                own_sensor.update_from_json(sensor_json_copy)
            else:
                sensor_json_copy.pop("_id", None)
                self.sensors.append(senseBoxSensor.from_json(sensor_json))
        self.id = d.get("_id")
        self.exposure = d.get("exposure")
        self.grouptag = d.get("grouptag")
        self.image = d.get("image")
        self.name = d.get("name")
        self.description = d.get("description")
        self.current_lat = lat
        self.current_lon = lon
        self.current_height = height
        self.weblink = d.get("weblink")

    @classmethod
    def from_json(cls, d):
        """
        Create a :any:`senseBox` from a :any:`dict`.

        Args:
            d (dict): the dictionary to create the :any:`senseBox` from

        Returns:
            senseBox : the senseBox
        """
        senseBox = cls()
        senseBox.update_from_json(d)
        return senseBox

    def to_json(self, with_id = True):
        """
        Serialize the metadata of this senseBox to a JSON dict for the API

        Args:
            with_id (bool, optional) : serialize the ids as well? Defaults to
                ``True``.

        Returns:
            dict : the JSON dict
        """
        box_json = {}
        if with_id:
            if self.id is not None:
                box_json["_id"] = self.id
        if self.name is not None:
            box_json["name"] = self.name
        if self.grouptag is not None:
            box_json["grouptag"] = self.grouptag
        box_json["location"] = {}
        if self.current_lat is not None:
            box_json["location"]["lat"] = float(self.current_lat)
        if self.current_lon is not None:
            box_json["location"]["lng"] = float(self.current_lon)
        if self.current_height is not None:
            box_json["location"]["height"] = float(self.current_height)
        if not "lat" in box_json["location"] or not "lng" in box_json["location"]:
            box_json.pop("location",None)
        if self.description is not None:
            box_json["description"] = self.description
        if self.weblink is not None:
            box_json["weblink"] = self.weblink
        if self.exposure is not None:
            box_json["exposure"] = self.exposure
        if self.sensors is not None:
            box_json["sensors"] = []
            for sensor in self.sensors:
                if sensor.id is None:
                    sensor_json = sensor.to_json(new = True)
                else:
                    if all([v for k,v in sensor.to_json().items()if k!="_id"]):
                        sensor_json = sensor.to_json(edited = True)
                    else:
                        sensor_json = sensor.to_json(deleted = True)
                box_json["sensors"].append(sensor_json)
        return box_json

    def upload_metadata(self):
        """
        Upload the metadata of this :any:`senseBox` to the API. Tries to call
        :any:`SenseMapAccount.update_box_metadata`.

        Returns:
            True : if the updating went well

        Raises:
            NoClientError : if this senseBox does not have an associated
                :any:`senseBoxClient` in :any:`senseBox.client`
        """
        if self.client is None:
            raise NoClientError("This box does not have an associated client.")
        if not hasattr(self.client, "update_box_metadata"):
            raise NoClientError("This box' client does is not able "
                "to update the metadata")
        return self.client.update_box_metadata(self)

    def upload_as_new(self):
        """
        Upload this :any:`senseBox` as a new senseBox. Tries to call
        :any:`SenseMapAccount.new_box`.

        Returns:
            True : if the updating went well

        Raises:
            NoClientError : if this senseBox does not have an associated
                :any:`senseBoxClient` in :any:`senseBox.client`
        """
        if self.client is None:
            raise NoClientError("This box does not have an associated client.")
        if not hasattr(self.client, "new_box"):
            raise NoClientError("This box' client does is not able "
                "to upload a new senseBox")
        return self.client.new_box(self)

    def upload_measurement(self, sensor):
        """
        Upload the current measurements of a given sensor.

        Args:
            sensor (senseBoxSensor): the sensor to upload data from

        Returns:
            True : if the updating went well

        Raises:
            NoClientError : if this senseBox does not have an associated
                :any:`senseBoxClient` in :any:`senseBox.client`
        """
        if self.client is None:
            raise NoClientError("This box does not have an associated client.")
        if not hasattr(self.client, "upload_current_measurement"):
            raise NoClientError("This box' client does is not able "
                "to upload measurements")
        return self.client.post_measurement(sensor)

    def fetch_metadata(self):
        """
        Fetch the metadata

        Raises:
            NoClientError : if this senseBox does not have an associated
                :any:`senseBoxClient` in :any:`senseBox.client`
        """
        if self.client is None:
            raise NoClientError("This box does not have an associated client.")
        if not hasattr(self.client, "_get_box"):
            raise NoClientError("This box' client does is not able "
                "to fetch boxes")
        assert self.id, "this senseBox does not have an id"
        box_json = self.client._get_box(self.id)
        self.update_from_json(box_json)

    def __eq__(self, other):
        """
        Compare this :any:`senseBox` to another :any:`senseBox` and check if
        all metadata is equal.
        """
        for arg in inspect.getfullargspec(self.__init__)[0][1:]:
            if not getattr(self, arg) == getattr(other, arg):
                return False
        return True



class senseBoxCollection(collections.abc.MutableSequence, ReprObject):
    """
    Container for multiple :any:`senseBox`s

    Args:
        boxes (sequence, optional): the :any:`senseBox`s
    """
    def __init__(self, boxes = []):
        self.boxes = boxes

    @property
    def boxes(self):
        """
        The underlying :any:`senseBox`es

        :type: any:`list`
        :setter: also adds the :any:`senseBoxCollection.client` to the
            given :any:`senseBox`s
        """
        try:
            self._boxes
        except AttributeError:
            self._boxes = []
        return self._boxes

    @boxes.setter
    def boxes(self, new_boxes):
        self._boxes = []
        for box in new_boxes:
            box.client = self.client
            self.append(box)

    @simplegetter
    def client(self):
        """
        The :any:`SenseMapClient` these :any:`senseBox`es belong to.

        :type: :any:`SenseMapClient` or :any:`None`
        """
        return None

    @simplesetter(client)
    def client(self, new):
        return new

    @property
    def by_id(self):
        """
        A :any:`dict` of these :any:`senseBox`s by their
        ids. boxes without an id get assigned to counter key
        ``new_sensor_N``.

        :type: :any:`dict`
        """
        i = itertools.count(0)
        return {
            box.id if box.id else ("new_box_{}".format(next(i))):\
                box for box in self}

    @property
    def by_name(self):
        """
        A :any:`dict` of these :any:`senseBox`es by their names. Boxes
        without a name get assigned to counter key ``unnamed box N``.
        Duplicate names get a ``" Nr. N"`` appended.

        :type: :any:`dict`
        """
        unnamed = itertools.count(0)
        d = {}
        duplicates = collections.Counter()
        for box in self:
            if box.name:
                if box.name in d:
                    duplicates[box.name] += 1
                    d["{} Nr. {}".format(box.name,
                        duplicates[box.name] + 1)] = box
                else:
                    d[box.name] = box
            else:
                d["Unnamed senseBox {}".format(next(unnamed))] = box
        return d

    def __getitem__(self, *args, **kwargs):
        return self.boxes.__getitem__(*args, **kwargs)

    def __setitem__(self, index, value):
        """
        Also adds the :any:`senseBoxCollection.client` to the new box
        """
        value.client = self.client
        return self.boxes.__setitem__(index, value)

    def insert(self, index, value):
        """
        Also adds the :any:`senseBoxCollection.client` to the new box
        """
        value.client = self.client
        return self.boxes.insert(index, value)

    def append(self, value):
        """
        Also adds the :any:`senseBoxCollection.client` to the new box
        """
        value.client = self.client
        return self.boxes.append(value)

    def __delitem__(self, *args, **kwargs):
        return self.boxes.__delitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        return self.boxes.__len__(*args, **kwargs)

    def __eq__(self, other):
        return self.boxes == other.boxes

    def __getattr__(self, attr):
        """
        Retreives a box by looking its title up in
        :any:`senseBoxCollection.by_name`.
        """
        try:
            return self.by_name[attr]
        except KeyError:
            raise AttributeError(
                "{} object does neither have an attribute '{}' "
                "nor a box with that title"
                .format(self.__class__.__name__, attr))




# system modules
import logging
import inspect
import json
import time
import urllib
import re
import functools

# internal modules
from sensemapi import paths
from sensemapi.client import SenseMapClient
from sensemapi.errors import *
from sensemapi.utils import *
from sensemapi.senseBox import senseBox, senseBoxCollection

# external modules
import requests

logger = logging.getLogger(__name__)

class SenseMapAccount(SenseMapClient):
    """
    Client to interface an `OpenSenseMap <https://opensensemap.org>`_ account
    via the `OpenSenseMap API <https://api.opensensemap.org>`_

    Args:
        email (str, optional): the user's login email address
        password (str, optional): the user's login password
    """
    def __init__(self,
        email = None,
        password = None,
        ):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @simplegetter
    def email(self):
        return None

    @simplesetter(email)
    def email(self, new):
        return new

    @simplegetter
    def password(self):
        return None

    @simplesetter(password)
    def password(self, new):
        return new

    @simplegetter
    def name(self):
        return None

    @simplesetter(name)
    def name(self, new):
        return new

    @simplegetter
    def role(self):
        return None

    @simplesetter(role)
    def role(self, new):
        return new

    @simplegetter
    def language(self):
        return None

    @simplesetter(language)
    def language(self, new):
        return new

    @property
    def boxes(self):
        try:
            self._boxes
        except AttributeError:
            self._boxes = senseBoxCollection()
            self._boxes.client = self
        return self._boxes

    @boxes.setter
    def boxes(self, new):
        if hasattr(new, "boxes"): # is already a senseBoxCollection
            self._boxes = new
        else:
            self._boxes = senseBoxCollection()
            self._boxes.client = self
            self._boxes.boxes = new

    @simplegetter
    def token(self):
        return None

    @simplesetter(token)
    def token(self, new):
        return new

    @simplegetter
    def refresh_token(self):
        return None

    @simplesetter(refresh_token)
    def refresh_token(self, new):
        return new

    @property
    def signed_in(self):
        """
        Whether the user is currently logged in

        :type: bool
        """
        return self.token is not None and self.refresh_token is not None

    @property
    def authorization_header(self):
        """
        Authorization header

        :type: :any:`dict`
        :getter: Return a :any:`requests`-compatible header-dict containing the
            authorization token. :any:`sign_in` if necessary.
        """
        if not self.token:
            self.sign_in()
        return {"Authorization": "Bearer {}".format(self.token)}

    def sign_in(self):
        """
        Sign in with the user's credentials.

        Returns:
            True : if the login process was successful

        Raises:
            OpenSenseMapAPIInvalidCredentialsError : wrong credentials
            OpenSenseMapAPIAuthenticationError : if the login did not work
        """
        assert self.email is not None, "No user email specified"
        assert self.password is not None, "No user password specified"
        response = requests.post(
            paths.SIGN_IN,
            json = {
                "email": self.email,
                "password": self.password,
                },
            )
        response_json = response.json()
        logger.debug("API responded JSON:\n{}"
            .format(pretty_json(response_json)))
        try:
            code = response_json["code"]
            message = response_json["message"]
            if code == "Forbidden":
                if re.search("user.+password.+not\s+valid",message.lower()):
                    raise OpenSenseMapAPIInvalidCredentialsError(message)
            if code != "Authorized":
                raise OpenSenseMapAPIAuthenticationError(
                    "Login did not work" + (": " + message or ""))
            self.token = response_json["token"]
            self.refresh_token = response_json["refreshToken"]
            data = response_json["data"]
            user = data["user"]
        except KeyError as e: # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "API response did not contain {}".format(e))
        self.read_user_details(user)
        return True

    def sign_out(self):
        """
        Sign out the user

        Returns:
            True : if the logout process was successful

        Raises:
            OpenSenseMapAPIOutdatedTokenError : if the logout did not work
        """
        if self.signed_in:
            response = requests.post(
                paths.SIGN_OUT,
                headers = self.authorization_header)
            response_json = response.json()
            logger.debug("API responded JSON:\n{}"
                .format(pretty_json(response_json)))
            try:
                code = response_json["code"]
                message = response_json["message"]
                if code == "Forbidden":
                    if re.search("invalid.+jwt",message.lower()):
                        raise OpenSenseMapAPIOutdatedTokenError(message)
                if code != "Ok":
                    raise OpenSenseMapAPIError(
                        "Logout did not work" + (": " + message or ""))
            except KeyError as e: # pragma: no cover
                raise OpenSenseMapAPIResponseError(
                    "API response did not contain {}".format(e))
        self.token = None
        self.refresh_token = None
        return True

    def refresh_tokens(self):
        """
        :any:`_refresh_tokens` and :any:`sign_in` again if necessary.

        Returns:
            True : if the refreshing was successful

        Raises:
            OpenSenseMapAPIAuthenticationError : if refreshing didn't work
        """
        try:
            self._refresh_tokens()
        except OpenSenseMapAPIOutdatedTokenError:
            logger.debug("Tokens are outdated. Signing in...")
            self.sign_in()
        return True

    def _refresh_tokens(self):
        """
        Attempt to refresh the tokens

        Returns:
            True : if the refreshing was successful

        Raises:
            OpenSenseMapAPIOutdatedTokenError : the refresh token is outdated
        """
        assert self.refresh_token, \
            "No refresh token available. Please sign_in()"
        response = requests.post(
            paths.REFRESH,
            headers = self.authorization_header,
            json = {"token": self.refresh_token},
            )
        response_json = response.json()
        logger.debug("API responded JSON:\n{}"
            .format(pretty_json(response_json)))
        try:
            code = response_json["code"]
            message = response_json["message"]
            if code == "Forbidden":
                if re.search("token.+invalid.+too.+old",message.lower()):
                    raise OpenSenseMapAPIOutdatedTokenError(message)
            if code != "Authorized":
                raise OpenSenseMapAPIError(
                    "Refreshing tokens did not work" + (": " + message or ""))
            self.token = response_json["token"]
            self.refresh_token = response_json["refreshToken"]
        except KeyError as e: # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "API response did not contain {}".format(e))
        return True

    def refresh_and_retry_on_outdated_tokens(decorated_function):
        """
        Decorator to call a decorated method, catching an
        :any:`OpenSenseMapAPIOutdatedTokenError`, then doing
        :any:`refresh_tokens` and then retries.
        """
        @functools.wraps(decorated_function)
        def wrapper(self, *args, **kwargs):
            try:
                return decorated_function(self, *args, **kwargs)
            except OpenSenseMapAPIOutdatedTokenError:
                logger.debug("Tokens are outdated. Refreshing tokens...")
                self.refresh_tokens()
            logger.debug("Now trying again to get details...")
            return decorated_function(self, *args, **kwargs)
        return wrapper

    def get_box(self, id):
        """
        Call :any:`SenseMapClient.get_box` and update own boxes if present

        Args:
            id (str) : the senseBox id to retreive

        Returns:
            senseBox : the retreived senseBox
        """
        box = super().get_box(id)
        try:
            own_box = self.boxes.by_id[box.id]
            own_box.update_from_json(box.to_json())
        except KeyError:
            pass
        return box

    def get_own_boxes(self):
        """
        Run :any:`get_details` to fetch the current box ids and then
        call :any:`get_box` on all of them.

        Returns:
            senseBoxCollection : all own boxes
        """
        self.get_details()
        for box in self.boxes:
            self.get_box(box.id)
        return self.boxes

    @refresh_and_retry_on_outdated_tokens
    def get_details(self):
        """
        Get the user details

        Returns:
            True : if the retrieval was successful

        Raises:
            OpenSenseMapAPIError : if the retreival did not work
        """
        response = requests.get(
            paths.USER_DETAILS,
            headers = self.authorization_header,
            )
        response_json = response.json()
        logger.debug("API responded JSON:\n{}"
            .format(pretty_json(response_json)))
        try:
            code = response_json["code"]
            message = response_json.get("message")
            if code == "Forbidden":
                if re.search("invalid.+jwt",message.lower()):
                    raise OpenSenseMapAPIOutdatedTokenError(
                        message or "Invalid JWT authorization")
            if code != "Ok":
                raise OpenSenseMapAPIError(
                    "Retrieving user details did not work" \
                        + (": " + message or ""))
            data = response_json["data"]
        except KeyError as e: # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "API response did not contain {}".format(e))
        try:
            user = data["me"]
        except KeyError as e: # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "API response data did not contain {}".format(e))
        self.read_user_details(user)
        return True

    @refresh_and_retry_on_outdated_tokens
    def update_box_metadata(self, box):
        """
        Update metadata of a :any:`senseBox`

        Args:
            box (senseBox) : the box to update

        Returns:
            True : if the updating went well
        """
        box_json = box.to_json(with_id = True)
        logger.debug("Uploading senseBox metadata:\n{}"
            .format(pretty_json(box_json)))
        response = requests.put(
            urllib.parse.urljoin(paths.BOXES+"/",box.id),
            headers = self.authorization_header,
            json = box_json,
            )
        response_json = response.json()
        logger.debug("API responded JSON:\n{}"
            .format(pretty_json(response_json)))
        try:
            code = response_json["code"]
            message = response_json.get("message")
            if code == "Forbidden":
                if re.search("invalid.+jwt",message.lower()):
                    raise OpenSenseMapAPIOutdatedTokenError(
                        message or "Invalid JWT authorization")
            if code != "Ok":
                raise OpenSenseMapAPIError(
                    "Could not update senseBox with id '{}'{}"
                    .format(box.id, ": {}".format(message) if message else ""))
            data = response_json["data"]
        except KeyError as e:
            raise OpenSenseMapAPIResponseError(
                "Response did not contain {}".format(e))
        # only keep edited boxes, not deleted and new ones
        edited_boxes_ids = set(sensor_json.get("_id") \
            for sensor_json in box_json.get("sensors",[]) \
                if (sensor_json.get("edited") and not sensor_json.get("new"))
            )
        box.sensors = [sensor for sensor in box.sensors \
            if sensor.id in edited_boxes_ids]
        # update the box data
        box.update_from_json(data)

    @refresh_and_retry_on_outdated_tokens
    def new_box(self, box):
        """
        Upload a new :any:`senseBox`

        Args:
            box (senseBox) : the box to upload

        Returns:
            senseBox : the responded new senseBox
        """
        assert box.exposure, "new box needs 'exposure'"
        assert box.current_lat, "new box needs 'current_lat'"
        assert box.current_lon, "new box needs 'current_lon'"
        assert box.name, "new box needs 'name'"
        assert box.sensors, "new box needs at least one sensor"
        box_json = box.to_json(with_id = False)
        logger.debug("Uploading new senseBox:\n{}"
            .format(pretty_json(box_json)))
        response = requests.post(
            paths.BOXES,
            headers = self.authorization_header,
            json = box_json,
            )
        response_json = response.json()
        logger.debug("API responded JSON:\n{}"
            .format(pretty_json(response_json)))
        try:
            code = response_json.get("code","")
            message = response_json.get("message","")
            if code == "Forbidden":
                if re.search("invalid.+jwt",message.lower()):
                    raise OpenSenseMapAPIOutdatedTokenError(
                        message or "Invalid JWT authorization")
            if not "successfully" in message:
                raise OpenSenseMapAPIError(
                    "Could not post new senseBox{}"
                        .format(": {}".format(message) if message else ""))
            data = response_json["data"]
        except KeyError as e:
            raise OpenSenseMapAPIResponseError(
                "Response did not contain {}".format(e))
        box.update_from_json(data)
        new_box = senseBox.from_json(data)
        logger.debug("new box:\n{}".format(new_box))
        self.get_details()
        return self.get_box(id = new_box.id)

    @refresh_and_retry_on_outdated_tokens
    def delete_box(self, box_id, really = False):
        """
        Mark a senseBox for deletion. This will also delete all the
        measurements. Deletion does not happen immediately. Also calls
        :any:`get_details`.

        Args:
            box_id (str) : the :any:`senseBox.id`
            really (bool, optional): really delete this box? Defaults to
                ``False``.

        Returns:
            True : if the marking for deletion went well
        """
        assert really, \
            "Refusing to delete box '{}' without really=True".format(box_id)
        response = requests.delete(
            urllib.parse.urljoin(paths.BOXES+"/",box_id),
            headers = self.authorization_header,
            json = {"password": self.password},
            )
        response_json = response.json()
        logger.debug("API responded JSON:\n{}"
            .format(pretty_json(response_json)))
        try:
            code = response_json.get("code","")
            message = response_json.get("message","")
            if code == "Forbidden":
                if re.search("invalid.+jwt",message.lower()):
                    raise OpenSenseMapAPIOutdatedTokenError(
                        message or "Invalid JWT authorization")
                if re.search("user.+not.+own.+box",message.lower()):
                    raise OpenSenseMapAPIPermissionError(
                        message or "User does not own box '{}'".format(box_id))
            if not response.status_code == 200:
                raise OpenSenseMapAPIError(
                    "Could not mark senseBox '{}' for deletion{}"
                    .format(box_id,": {}".format(message) if message else ""))
        except KeyError as e:
            raise OpenSenseMapAPIResponseError(
                "Response did not contain {}".format(e))
        self.get_details()
        return True

    @refresh_and_retry_on_outdated_tokens
    def upload_current_measurement(self, box, sensor):
        """
        Upload the current measurements of a given sensor of a given senseBox

        Args:
            box_id (str) : the :any:`senseBox.id`
            really (bool, optional): really delete this box? Defaults to
                ``False``.

        Returns:
            True : if the upload went well
        """
        assert box.id, "box has no 'id'"
        assert sensor.id, "sensor has no 'id'"
        response = requests.post(
            urllib.parse.urljoin(paths.BOXES+"/",box.id+"/",sensor.id),
            headers = self.authorization_header,
            json = {},
            )
        response_json = response.json()
        logger.debug("API responded JSON:\n{}"
            .format(pretty_json(response_json)))
        try:
            code = response_json.get("code","")
            message = response_json.get("message","")
            if code == "Forbidden":
                if re.search("invalid.+jwt",message.lower()):
                    raise OpenSenseMapAPIOutdatedTokenError(
                        message or "Invalid JWT authorization")
                if re.search("user.+not.+own.+box",message.lower()):
                    raise OpenSenseMapAPIPermissionError(
                        message or "User does not own box '{}'".format(box_id))
            if not response.status_code == 200:
                raise OpenSenseMapAPIError(
                    "Could not mark senseBox '{}' for deletion{}"
                    .format(box_id,": {}".format(message) if message else ""))
        except KeyError as e:
            raise OpenSenseMapAPIResponseError(
                "Response did not contain {}".format(e))
        self.get_details()
        return True

    def read_user_details(self, user_details, exception=OpenSenseMapAPIError):
        """
        Read user details into object properties

        Args:
            user_details (dict): the user details as returned by the API
            exception (Exception): the exception to raise if the data is
                ill-formed

        Returns:
            True : if everything is alright

        Raises:
            ``exception`` : if the data is ill-formed
        """
        try:
            email = user_details["email"]
            if email != self.email:
                logger.info("The user's email '{}' is not the same "
                    "as the login email '{}'?!?".format(self.email, email))
            if not user_details["emailIsConfirmed"]:
                logger.warning("The email '{}' is unconfirmed!".format(email))
            self.email = user_details["email"]
            self.name = user_details["name"]
            self.role = user_details["role"]
            self.language = user_details["language"]
            for own_box in self.boxes:
                if not own_box.id in user_details["boxes"]:
                    self.boxes.remove(own_box)
            for box_id in user_details["boxes"]:
                if not box_id in self.boxes.by_id:
                    box = senseBox(id = box_id)
                    self.boxes.append(box)
        except KeyError as e: # pragma: no cover
            raise exception(
                "user information did not contain {}".format(e))
        return True

    def __del__(self):
        """
        Class deconstructor. Calls :any:`sign_out`.
        """
        self.sign_out()
